@@ -30,10 +30,11 @@
 #endif
 
 #include "system.h"
+#include "acl.h"
 #include "backupfile.h"
+#include "buffer-lcm.h"
 #include "copy.h"
 #include "cp-hash.h"
-#include "dirname.h"
 #include "euidaccess.h"
 #include "error.h"
 #include "fcntl--.h"
@@ -42,6 +43,7 @@
 #include "getpagesize.h"
 #include "hash.h"
 #include "hash-pjw.h"
+#include "lchmod.h"
 #include "quote.h"
 #include "same.h"
 #include "savedir.h"
@@ -51,11 +53,9 @@
 #include "xreadlink.h"
 #include "yesno.h"
 
-#ifndef HAVE_FCHMOD
-# define HAVE_FCHMOD false
-#endif
 #ifndef HAVE_FCHOWN
 # define HAVE_FCHOWN false
+# define fchown(fd, uid, gid) (-1)
 #endif
 
 #define SAME_OWNER(A, B) ((A).st_uid == (B).st_uid)
@@ -101,26 +101,6 @@ static char const *top_level_dst_name;
 /* The invocation name of this program.  */
 extern char *program_name;
 
-/* Encapsulate selection of the file mode to be applied to
-   new non-directories.  */
-
-static mode_t
-get_dest_mode (const struct cp_options *option, mode_t mode)
-{
-  /* In some applications (e.g., install), use precisely the
-     specified mode.  */
-  if (option->set_mode)
-    return option->mode;
-
-  /* Honor the umask for `cp', but not for `mv' or `cp -p'.
-     In addition, `cp' without -p must clear the set-user-ID and set-group-ID
-     bits.  POSIX requires it do that when creating new files.  */
-  if (!option->move_mode && !option->preserve_mode)
-    mode &= (option->umask_kill & ~(S_ISUID | S_ISGID));
-
-  return mode;
-}
-
 /* FIXME: describe */
 /* FIXME: rewrite this to use a hash table so we avoid the quadratic
    performance hit that's probably noticeable only on trees deeper
@@ -191,6 +171,65 @@ copy_dir (char const *src_name_in, char const *dst_name_in, bool new_dst,
   return ok;
 }
 
+/* Set the owner and owning group of DEST_DESC to the st_uid and
+   st_gid fields of SRC_SB.  If DEST_DESC is undefined (-1), set
+   the owner and owning group of DST_NAME instead.  DEST_DESC must
+   refer to the same file as DEST_NAME if defined.
+   Return true if the syscall succeeds, or if it's ok not to
+   preserve ownership.  */
+
+static bool
+set_owner (const struct cp_options *x, char const *dst_name, int dest_desc,
+	   uid_t uid, gid_t gid)
+{
+  if (HAVE_FCHOWN && dest_desc != -1)
+    {
+      if (fchown (dest_desc, uid, gid) == 0)
+	return true;
+    }
+  else
+    {
+      if (chown (dst_name, uid, gid) == 0)
+	return true;
+    }
+
+  if (! chown_failure_ok (x))
+    {
+      error (0, errno, _("failed to preserve ownership for %s"),
+	     quote (dst_name));
+      if (x->require_preserve)
+	return false;
+    }
+
+  return true;
+}
+
+/* Set the st_author field of DEST_DESC to the st_author field of
+   SRC_SB. If DEST_DESC is undefined (-1), set the st_author field
+   of DST_NAME instead.  DEST_DESC must refer to the same file as
+   DEST_NAME if defined.  */
+
+static void
+set_author (const char *dst_name, int dest_desc, const struct stat *src_sb)
+{
+#if HAVE_STRUCT_STAT_ST_AUTHOR
+  /* Preserve the st_author field.  */
+  file_t file = (dest_desc < 0
+		 ? file_name_lookup (dst_name, 0, 0)
+		 : getdport (dest_desc));
+  if (file == MACH_PORT_NULL)
+    error (0, errno, _("failed to lookup file %s"), quote (dst_name));
+  else
+    {
+      error_t err = file_chauthor (file, src_sb->st_author);
+      if (err)
+	error (0, err, _("failed to preserve authorship for %s"),
+	       quote (dst_name));
+      mach_port_deallocate (mach_task_self (), file);
+    }
+#endif
+}
+
 /* Copy a regular file from SRC_NAME to DST_NAME.
    If the source file contains holes, copies holes and blocks of zeros
    in the source file as holes in the destination file.
@@ -198,30 +237,21 @@ copy_dir (char const *src_name_in, char const *dst_name_in, bool new_dst,
    Use DST_MODE as the 3rd argument in the call to open.
    X provides many option settings.
    Return true if successful.
-   *NEW_DST and *CHOWN_SUCCEEDED are as in copy_internal.
-   SRC_SB and DST_SB are the results of calling XSTAT (aka stat for
-   SRC_SB) on SRC_NAME and DST_NAME.  */
+   *NEW_DST is as in copy_internal.
+   SRC_SB is the result of calling XSTAT (aka stat) on SRC_NAME.  */
 
 static bool
 copy_reg (char const *src_name, char const *dst_name,
 	  const struct cp_options *x, mode_t dst_mode, bool *new_dst,
-	  bool *chown_succeeded,
-	  struct stat const *src_sb,
-	  struct stat const *dst_sb)
+	  struct stat const *src_sb)
 {
   char *buf;
-  size_t buf_size;
-  size_t buf_alignment;
+  char *buf_alloc = NULL;
   int dest_desc;
   int source_desc;
   struct stat sb;
   struct stat src_open_sb;
-  char *cp;
-  int *ip;
   bool return_val = true;
-  off_t n_read_total = 0;
-  bool last_write_made_hole = false;
-  bool make_holes = false;
 
   source_desc = open (src_name, O_RDONLY | O_BINARY);
   if (source_desc < 0)
@@ -282,8 +312,6 @@ copy_reg (char const *src_name, char const *dst_name,
       goto close_src_desc;
     }
 
-  /* Determine the optimal buffer size.  */
-
   if (fstat (dest_desc, &sb))
     {
       error (0, errno, _("cannot fstat %s"), quote (dst_name));
@@ -291,126 +319,167 @@ copy_reg (char const *src_name, char const *dst_name,
       goto close_src_and_dst_desc;
     }
 
-  buf_size = ST_BLKSIZE (sb);
+  if (! (S_ISREG (src_open_sb.st_mode) && src_open_sb.st_size == 0))
+    {
+      typedef uintptr_t word;
+      off_t n_read_total = 0;
 
-  /* Even with --sparse=always, try to create holes only
-     if the destination is a regular file.  */
-  if (x->sparse_mode == SPARSE_ALWAYS && S_ISREG (sb.st_mode))
-    make_holes = true;
+      /* Choose a suitable buffer size; it may be adjusted later.  */
+      size_t buf_alignment = lcm (getpagesize (), sizeof (word));
+      size_t buf_alignment_slop = sizeof (word) + buf_alignment - 1;
+      size_t buf_size = ST_BLKSIZE (sb);
+
+      /* Deal with sparse files.  */
+      bool last_write_made_hole = false;
+      bool make_holes = false;
+
+      if (S_ISREG (sb.st_mode))
+	{
+	  /* Even with --sparse=always, try to create holes only
+	     if the destination is a regular file.  */
+	  if (x->sparse_mode == SPARSE_ALWAYS)
+	    make_holes = true;
 
 #if HAVE_STRUCT_STAT_ST_BLOCKS
-  if (x->sparse_mode == SPARSE_AUTO && S_ISREG (sb.st_mode))
-    {
-      /* Use a heuristic to determine whether SRC_NAME contains any
-	 sparse blocks. */
-
-      if (fstat (source_desc, &sb))
-	{
-	  error (0, errno, _("cannot fstat %s"), quote (src_name));
-	  return_val = false;
-	  goto close_src_and_dst_desc;
+	  /* Use a heuristic to determine whether SRC_NAME contains any sparse
+	     blocks.  If the file has fewer blocks than would normally be
+	     needed for a file of its size, then at least one of the blocks in
+	     the file is a hole.  */
+	  if (x->sparse_mode == SPARSE_AUTO && S_ISREG (src_open_sb.st_mode)
+	      && ST_NBLOCKS (src_open_sb) < src_open_sb.st_size / ST_NBLOCKSIZE)
+	    make_holes = true;
+#endif
 	}
 
-      /* If the file has fewer blocks than would normally
-	 be needed for a file of its size, then
-	 at least one of the blocks in the file is a hole. */
-      if (S_ISREG (sb.st_mode)
-	  && sb.st_size / ST_NBLOCKSIZE > ST_NBLOCKS (sb))
-	make_holes = true;
-    }
-#endif
-
-  /* Make a buffer with space for a sentinel at the end.  */
-
-  buf_alignment = lcm (getpagesize (), sizeof (int));
-  buf = alloca (buf_size + sizeof (int) + buf_alignment - 1);
-  buf = ptr_align (buf, buf_alignment);
-
-  for (;;)
-    {
-      ssize_t n_read = read (source_desc, buf, buf_size);
-      if (n_read < 0)
+      /* If not making a sparse file, try to use a more-efficient
+	 buffer size.  */
+      if (! make_holes)
 	{
-#ifdef EINTR
-	  if (errno == EINTR)
-	    continue;
-#endif
-	  error (0, errno, _("reading %s"), quote (src_name));
-	  return_val = false;
-	  goto close_src_and_dst_desc;
+	  /* These days there's no point ever messing with buffers smaller
+	     than 8 KiB.  It would be nice to configure SMALL_BUF_SIZE
+	     dynamically for this host and pair of files, but there doesn't
+	     seem to be a good way to get readahead info portably.  */
+	  enum { SMALL_BUF_SIZE = 8 * 1024 };
+
+	  /* Compute the least common multiple of the input and output
+	     buffer sizes, adjusting for outlandish values.  */
+	  size_t blcm_max = MIN (SIZE_MAX, SSIZE_MAX) - buf_alignment_slop;
+	  size_t blcm = buffer_lcm (ST_BLKSIZE (src_open_sb), buf_size,
+				    blcm_max);
+
+	  /* Do not use a block size that is too small.  */
+	  buf_size = MAX (SMALL_BUF_SIZE, blcm);
+
+	  /* Do not bother with a buffer larger than the input file, plus one
+	     byte to make sure the file has not grown while reading it.  */
+	  if (S_ISREG (src_open_sb.st_mode) && src_open_sb.st_size < buf_size)
+	    buf_size = src_open_sb.st_size + 1;
+
+	  /* However, stick with a block size that is a positive multiple of
+	     blcm, overriding the above adjustments.  Watch out for
+	     overflow.  */
+	  buf_size += blcm - 1;
+	  buf_size -= buf_size % blcm;
+	  if (buf_size == 0 || blcm_max < buf_size)
+	    buf_size = blcm;
 	}
-      if (n_read == 0)
-	break;
 
-      n_read_total += n_read;
+      /* Make a buffer with space for a sentinel at the end.  */
+      buf_alloc = xmalloc (buf_size + buf_alignment_slop);
+      buf = ptr_align (buf_alloc, buf_alignment);
 
-      ip = NULL;
-      if (make_holes)
+      for (;;)
 	{
-	  buf[n_read] = 1;	/* Sentinel to stop loop.  */
+	  word *wp = NULL;
 
-	  /* Find first nonzero *word*, or the word with the sentinel.  */
-
-	  ip = (int *) buf;
-	  while (*ip++ == 0)
-	    ;
-
-	  /* Find the first nonzero *byte*, or the sentinel.  */
-
-	  cp = (char *) (ip - 1);
-	  while (*cp++ == 0)
-	    ;
-
-	  /* If we found the sentinel, the whole input block was zero,
-	     and we can make a hole.  */
-
-	  if (cp > buf + n_read)
+	  ssize_t n_read = read (source_desc, buf, buf_size);
+	  if (n_read < 0)
 	    {
-	      /* Make a hole.  */
-	      if (lseek (dest_desc, (off_t) n_read, SEEK_CUR) < 0L)
+#ifdef EINTR
+	      if (errno == EINTR)
+		continue;
+#endif
+	      error (0, errno, _("reading %s"), quote (src_name));
+	      return_val = false;
+	      goto close_src_and_dst_desc;
+	    }
+	  if (n_read == 0)
+	    break;
+
+	  n_read_total += n_read;
+
+	  if (make_holes)
+	    {
+	      char *cp;
+
+	      buf[n_read] = 1;	/* Sentinel to stop loop.  */
+
+	      /* Find first nonzero *word*, or the word with the sentinel.  */
+
+	      wp = (word *) buf;
+	      while (*wp++ == 0)
+		continue;
+
+	      /* Find the first nonzero *byte*, or the sentinel.  */
+
+	      cp = (char *) (wp - 1);
+	      while (*cp++ == 0)
+		continue;
+
+	      if (cp <= buf + n_read)
+		/* Clear to indicate that a normal write is needed. */
+		wp = NULL;
+	      else
 		{
-		  error (0, errno, _("cannot lseek %s"), quote (dst_name));
+		  /* We found the sentinel, so the whole input block was zero.
+		     Make a hole.  */
+		  if (lseek (dest_desc, n_read, SEEK_CUR) < 0)
+		    {
+		      error (0, errno, _("cannot lseek %s"), quote (dst_name));
+		      return_val = false;
+		      goto close_src_and_dst_desc;
+		    }
+		  last_write_made_hole = true;
+		}
+	    }
+
+	  if (!wp)
+	    {
+	      size_t n = n_read;
+	      if (full_write (dest_desc, buf, n) != n)
+		{
+		  error (0, errno, _("writing %s"), quote (dst_name));
 		  return_val = false;
 		  goto close_src_and_dst_desc;
 		}
-	      last_write_made_hole = true;
+	      last_write_made_hole = false;
+
+	      /* A short read on a regular file means EOF.  */
+	      if (n_read != buf_size && S_ISREG (src_open_sb.st_mode))
+		break;
 	    }
-	  else
-	    /* Clear to indicate that a normal write is needed. */
-	    ip = NULL;
 	}
-      if (ip == NULL)
+
+      /* If the file ends with a `hole', we need to do something to record
+	 the length of the file.  On modern systems, calling ftruncate does
+	 the job.  On systems without native ftruncate support, we have to
+	 write a byte at the ending position.  Otherwise the kernel would
+	 truncate the file at the end of the last write operation.  */
+
+      if (last_write_made_hole)
 	{
-	  size_t n = n_read;
-	  if (full_write (dest_desc, buf, n) != n)
+	  if (HAVE_FTRUNCATE
+	      ? /* ftruncate sets the file size,
+		   so there is no need for a write.  */
+	      ftruncate (dest_desc, n_read_total) < 0
+	      : /* Seek backwards one character and write a null.  */
+	      (lseek (dest_desc, (off_t) -1, SEEK_CUR) < 0L
+	       || full_write (dest_desc, "", 1) != 1))
 	    {
 	      error (0, errno, _("writing %s"), quote (dst_name));
 	      return_val = false;
 	      goto close_src_and_dst_desc;
 	    }
-	  last_write_made_hole = false;
-	}
-    }
-
-  /* If the file ends with a `hole', something needs to be written at
-     the end.  Otherwise the kernel would truncate the file at the end
-     of the last write operation.  */
-
-  if (last_write_made_hole)
-    {
-#if HAVE_FTRUNCATE
-      /* Write a null character and truncate it again.  */
-      if (full_write (dest_desc, "", 1) != 1
-	  || ftruncate (dest_desc, n_read_total) < 0)
-#else
-      /* Seek backwards one character and write a null.  */
-      if (lseek (dest_desc, (off_t) -1, SEEK_CUR) < 0L
-	  || full_write (dest_desc, "", 1) != 1)
-#endif
-	{
-	  error (0, errno, _("writing %s"), quote (dst_name));
-	  return_val = false;
-	  goto close_src_and_dst_desc;
 	}
     }
 
@@ -431,49 +500,28 @@ copy_reg (char const *src_name, char const *dst_name,
 	}
     }
 
-#if HAVE_FCHOWN
-  if (x->preserve_ownership
-      && (*new_dst || !SAME_OWNER_AND_GROUP (*src_sb, *dst_sb)))
+  if (x->preserve_ownership && ! SAME_OWNER_AND_GROUP (*src_sb, sb))
     {
-      if (fchown (dest_desc, src_sb->st_uid, src_sb->st_gid) == 0)
-	*chown_succeeded = true;
-      else if (! chown_failure_ok (x))
-	{
-	  error (0, errno, _("failed to preserve ownership for %s"),
-		 quote (dst_name));
-	  if (x->require_preserve)
-	    {
-	      return_val = false;
-	      goto close_src_and_dst_desc;
-	    }
+      if (! set_owner (x, dst_name, dest_desc, src_sb->st_uid, src_sb->st_gid))
+        {
+	  return_val = false;
+	  goto close_src_and_dst_desc;
 	}
     }
-#endif
 
-#if HAVE_STRUCT_STAT_ST_AUTHOR
-  /* FIXME: Preserve the st_author field via the file descriptor dest_desc.  */
-#endif
+  set_author (dst_name, dest_desc, src_sb);
 
-#if HAVE_FCHMOD
-  /* Permissions of newly-created regular files were set upon `open'.
-     But don't return early if there were any special bits and chown
-     succeeded, because the chown must have reset those bits.  */
-  if (!(*new_dst
-	&& !(*chown_succeeded && (src_sb->st_mode & ~S_IRWXUGO)))
-      && (x->preserve_mode || *new_dst)
-      && (x->copy_as_regular || S_ISREG (src_sb->st_mode)))
+  if (x->preserve_mode || x->move_mode)
     {
-      if (fchmod (dest_desc, get_dest_mode (x, src_sb->st_mode)) != 0)
-	{
-	  error (0, errno, _("setting permissions for %s"), quote (dst_name));
-	  if (x->set_mode || x->require_preserve)
-	    {
-	      return_val = false;
-	      goto close_src_and_dst_desc;
-	    }
-	}
+      if (copy_acl (src_name, source_desc, dst_name, dest_desc,
+		    src_sb->st_mode) != 0 && x->require_preserve)
+	return_val = false;
+    }
+  else if (x->set_mode)
+    {
+      if (set_acl (dst_name, dest_desc, x->mode) != 0)
+	return_val = false;
     }
-#endif
 
 close_src_and_dst_desc:
   if (close (dest_desc) < 0)
@@ -488,6 +536,7 @@ close_src_desc:
       return_val = false;
     }
 
+  free (buf_alloc);
   return return_val;
 }
 
@@ -540,8 +589,8 @@ same_file_ok (char const *src_name, struct stat const *src_sb,
       same_link = same;
 
       /* If both the source and destination files are symlinks (and we'll
-	 know this here IFF preserving symlinks (aka xstat == lstat),
-	 then it's ok -- as long as they are distinct.  */
+	 know this here IFF preserving symlinks), then it's ok -- as long
+	 as they are distinct.  */
       if (S_ISLNK (src_sb->st_mode) && S_ISLNK (dst_sb->st_mode))
 	return ! same_name (src_name, dst_name);
 
@@ -883,6 +932,18 @@ abandon_move (const struct cp_options *x,
               && ! yesno ()));
 }
 
+/* Print --verbose output on standard output, e.g. `new' -> `old'.
+   If BACKUP_DST_NAME is non-NULL, then also indicate that it is
+   the name of a backup file.  */
+static void
+emit_verbose (char const *src, char const *dst, char const *backup_dst_name)
+{
+  printf ("%s -> %s", quote_n (0, src), quote_n (1, dst));
+  if (backup_dst_name)
+    printf (_(" (backup: %s)"), quote (backup_dst_name));
+  putchar ('\n');
+}
+
 /* Copy the file SRC_NAME to the file DST_NAME.  The files may be of
    any type.  NEW_DST should be true if the file DST_NAME cannot
    exist because its parent directory was just created; NEW_DST should
@@ -894,7 +955,6 @@ abandon_move (const struct cp_options *x,
    Set *COPY_INTO_SELF if SRC_NAME is a parent of (or the
    same as) DST_NAME; otherwise, clear it.
    Return true if successful.  */
-
 static bool
 copy_internal (char const *src_name, char const *dst_name,
 	       bool new_dst,
@@ -909,12 +969,13 @@ copy_internal (char const *src_name, char const *dst_name,
   struct stat dst_sb;
   mode_t src_mode;
   mode_t src_type;
+  mode_t dst_mode IF_LINT (= 0);
+  bool restore_dst_mode = false;
   char *earlier_file = NULL;
   char *dst_backup = NULL;
   bool backup_succeeded = false;
   bool delayed_ok;
   bool copied_as_regular = false;
-  bool chown_succeeded = false;
   bool preserve_metadata;
 
   if (x->move_mode && rename_succeeded)
@@ -971,7 +1032,8 @@ copy_internal (char const *src_name, char const *dst_name,
 	    }
 	}
       else
-	{
+	{ /* Here, we know that dst_name exists, at least to the point
+	     that it is XSTAT'able.  */
 	  bool return_now;
 	  bool unlink_src;
 	  bool ok = same_file_ok (src_name, &src_sb, dst_name, &dst_sb,
@@ -1003,10 +1065,18 @@ copy_internal (char const *src_name, char const *dst_name,
 	    {
 	      if (S_ISDIR (src_type))
 		{
-		  error (0, 0,
-		     _("cannot overwrite non-directory %s with directory %s"),
-			 quote_n (0, dst_name), quote_n (1, src_name));
-		  return false;
+		  if (x->move_mode && x->backup_type != no_backups)
+		    {
+		      /* Moving a directory onto an existing
+			 non-directory is ok only with --backup.  */
+		    }
+		  else
+		    {
+		      error (0, 0,
+		       _("cannot overwrite non-directory %s with directory %s"),
+			     quote_n (0, dst_name), quote_n (1, src_name));
+		      return false;
+		    }
 		}
 
 	      /* Don't let the user destroy their data, even if they try hard:
@@ -1031,10 +1101,18 @@ copy_internal (char const *src_name, char const *dst_name,
 	    {
 	      if (S_ISDIR (dst_sb.st_mode))
 		{
-		  error (0, 0,
-		       _("cannot overwrite directory %s with non-directory"),
-			 quote (dst_name));
-		  return false;
+		  if (x->move_mode && x->backup_type != no_backups)
+		    {
+		      /* Moving a non-directory onto an existing
+			 directory is ok only with --backup.  */
+		    }
+		  else
+		    {
+		      error (0, 0,
+			 _("cannot overwrite directory %s with non-directory"),
+			     quote (dst_name));
+		      return false;
+		    }
 		}
 
 	      if (x->update)
@@ -1096,7 +1174,8 @@ copy_internal (char const *src_name, char const *dst_name,
 	  if (x->move_mode)
 	    {
 	      /* Don't allow user to move a directory onto a non-directory.  */
-	      if (S_ISDIR (src_sb.st_mode) && !S_ISDIR (dst_sb.st_mode))
+	      if (S_ISDIR (src_sb.st_mode) && !S_ISDIR (dst_sb.st_mode)
+		  && x->backup_type == no_backups)
 		{
 		  error (0, 0,
 		       _("cannot move directory onto non-directory: %s -> %s"),
@@ -1105,7 +1184,10 @@ copy_internal (char const *src_name, char const *dst_name,
 		}
 	    }
 
-	  if (x->backup_type != no_backups && !S_ISDIR (dst_sb.st_mode))
+	  if (x->backup_type != no_backups
+	      /* Don't try to back up a destination if the last
+		 component of src_name is "." or "..".  */
+	      && ! dot_or_dotdot (last_component (src_name)))
 	    {
 	      char *tmp_backup = find_backup_file_name (dst_name,
 							x->backup_type);
@@ -1176,12 +1258,7 @@ copy_internal (char const *src_name, char const *dst_name,
      directory.  So --verbose should not announce anything until we're
      sure we'll create a directory. */
   if (x->verbose && !S_ISDIR (src_type))
-    {
-      printf ("%s -> %s", quote_n (0, src_name), quote_n (1, dst_name));
-      if (backup_succeeded)
-	printf (_(" (backup: %s)"), quote (dst_backup));
-      putchar ('\n');
-    }
+    emit_verbose (src_name, dst_name, backup_succeeded ? dst_backup : NULL);
 
   /* Associate the destination file name with the source device and inode
      so that if we encounter a matching dev/ino pair in the source tree
@@ -1296,7 +1373,9 @@ copy_internal (char const *src_name, char const *dst_name,
       if (rename (src_name, dst_name) == 0)
 	{
 	  if (x->verbose && S_ISDIR (src_type))
-	    printf ("%s -> %s\n", quote_n (0, src_name), quote_n (1, dst_name));
+	    emit_verbose (src_name, dst_name,
+			  backup_succeeded ? dst_backup : NULL);
+
 	  if (rename_succeeded)
 	    *rename_succeeded = true;
 
@@ -1320,18 +1399,7 @@ copy_internal (char const *src_name, char const *dst_name,
 
       /* This happens when attempting to rename a directory to a
 	 subdirectory of itself.  */
-      if (errno == EINVAL
-
-	  /* When src_name is on an NFS file system, some types of
-	     clients, e.g., SunOS4.1.4 and IRIX-5.3, set errno to EIO
-	     instead.  Testing for this here risks misinterpreting a real
-	     I/O error as an attempt to move a directory into itself, so
-	     FIXME: consider not doing this.  */
-	  || errno == EIO
-
-	  /* And with SunOS-4.1.4 client and OpenBSD-2.3 server,
-	     we get ENOTEMPTY.  */
-	  || errno == ENOTEMPTY)
+      if (errno == EINVAL)
 	{
 	  /* FIXME: this is a little fragile in that it relies on rename(2)
 	     failing with a specific errno value.  Expect problems on
@@ -1369,7 +1437,7 @@ copy_internal (char const *src_name, char const *dst_name,
       if (errno != EXDEV)
 	{
 	  /* There are many ways this can happen due to a race condition.
-	     When something happens between the initial xstat and the
+	     When something happens between the initial XSTAT and the
 	     subsequent rename, we can get many different types of errors.
 	     For example, if the destination is initially a non-directory
 	     or non-existent, but it is created as a directory, the rename
@@ -1433,25 +1501,45 @@ copy_internal (char const *src_name, char const *dst_name,
 
       if (new_dst || !S_ISDIR (dst_sb.st_mode))
 	{
-	  /* Create the new directory writable and searchable, so
-             we can create new entries in it.  */
-
-	  if (mkdir (dst_name, (src_mode & x->umask_kill) | S_IRWXU) != 0)
+	  if (mkdir (dst_name, src_mode) != 0)
 	    {
 	      error (0, errno, _("cannot create directory %s"),
 		     quote (dst_name));
 	      goto un_backup;
 	    }
 
+	  /* We need search and write permissions to the new directory
+	     for writing the directory's contents. Check if these
+	     permissions are there.  */
+
+	  if (lstat (dst_name, &dst_sb) != 0)
+	    {
+	      error (0, errno, _("cannot stat %s"), quote (dst_name));
+	      goto un_backup;
+	    }
+	  else if ((dst_sb.st_mode & S_IRWXU) != S_IRWXU)
+	    {
+	      /* Make the new directory searchable and writable.  */
+
+	      dst_mode = dst_sb.st_mode;
+	      restore_dst_mode = true;
+
+	      if (lchmod (dst_name, dst_mode | S_IRWXU) != 0)
+		{
+		  error (0, errno, _("setting permissions for %s"),
+			 quote (dst_name));
+		  goto un_backup;
+		}
+	    }
+
 	  /* Insert the created directory's inode and device
              numbers into the search structure, so that we can
              avoid copying it again.  */
 
-	  if (! remember_created (dst_name))
-	    goto un_backup;
+	  remember_copied (dst_name, dst_sb.st_ino, dst_sb.st_dev);
 
 	  if (x->verbose)
-	    printf ("%s -> %s\n", quote_n (0, src_name), quote_n (1, dst_name));
+	    emit_verbose (src_name, dst_name, NULL);
 	}
 
       /* Are we crossing a file system boundary?  */
@@ -1469,7 +1557,6 @@ copy_internal (char const *src_name, char const *dst_name,
 	  delayed_ok = false;
 	}
     }
-#ifdef S_ISLNK
   else if (x->symbolic_link)
     {
       preserve_metadata = false;
@@ -1508,8 +1595,22 @@ copy_internal (char const *src_name, char const *dst_name,
 	  goto un_backup;
 	}
     }
+
+  else if (x->hard_link
+#ifdef LINK_FOLLOWS_SYMLINKS
+  /* A POSIX-conforming link syscall dereferences a symlink, yet cp,
+     invoked with `--link --no-dereference', should not.  Thus, with
+     a POSIX-conforming link system call, we can't use link() here,
+     since that would create a hard link to the referent (effectively
+     dereferencing the symlink), rather than to the symlink itself.
+     We can approximate the desired behavior by skipping this hard-link
+     creating block and instead copying the symlink, via the `S_ISLNK'-
+     copying code below.
+     When link operates on the symlinks themselves, we use this block
+     and just call link().  */
+	   && !(S_ISLNK (src_mode) && x->dereference == DEREF_NEVER)
 #endif
-  else if (x->hard_link)
+	   )
     {
       preserve_metadata = false;
       if (link (src_name, dst_name))
@@ -1525,36 +1626,27 @@ copy_internal (char const *src_name, char const *dst_name,
       /* POSIX says the permission bits of the source file must be
 	 used as the 3rd argument in the open call, but that's not consistent
 	 with historical practice.  */
-      if (! copy_reg (src_name, dst_name, x,
-		      get_dest_mode (x, src_mode), &new_dst, &chown_succeeded,
-		      &src_sb, &dst_sb))
+      if (! copy_reg (src_name, dst_name, x, src_mode, &new_dst, &src_sb))
 	goto un_backup;
     }
-  else
-#ifdef S_ISFIFO
-  if (S_ISFIFO (src_type))
+  else if (S_ISFIFO (src_type))
     {
-      if (mkfifo (dst_name, get_dest_mode (x, src_mode)))
+      if (mkfifo (dst_name, src_mode))
 	{
 	  error (0, errno, _("cannot create fifo %s"), quote (dst_name));
 	  goto un_backup;
 	}
     }
-  else
-#endif
-    if (S_ISBLK (src_type) || S_ISCHR (src_type)
-	|| S_ISSOCK (src_type))
+  else if (S_ISBLK (src_type) || S_ISCHR (src_type) || S_ISSOCK (src_type))
     {
-      if (mknod (dst_name, get_dest_mode (x, src_mode), src_sb.st_rdev))
+      if (mknod (dst_name, src_mode, src_sb.st_rdev))
 	{
 	  error (0, errno, _("cannot create special file %s"),
 		 quote (dst_name));
 	  goto un_backup;
 	}
     }
-  else
-#ifdef S_ISLNK
-  if (S_ISLNK (src_type))
+  else if (S_ISLNK (src_type))
     {
       char *src_link_val = xreadlink (src_name, src_sb.st_size);
       if (src_link_val == NULL)
@@ -1598,7 +1690,7 @@ copy_internal (char const *src_name, char const *dst_name,
 	{
 	  /* Preserve the owner and group of the just-`copied'
 	     symbolic link, if possible.  */
-# if HAVE_LCHOWN
+#if HAVE_LCHOWN
 	  if (lchown (dst_name, src_sb.st_uid, src_sb.st_gid) != 0
 	      && ! chown_failure_ok (x))
 	    {
@@ -1606,16 +1698,15 @@ copy_internal (char const *src_name, char const *dst_name,
 		     dst_name);
 	      goto un_backup;
 	    }
-# else
+#else
 	  /* Can't preserve ownership of symlinks.
 	     FIXME: maybe give a warning or even error for symlinks
 	     in directories with the sticky bit set -- there, not
 	     preserving owner/group is a potential security problem.  */
-# endif
+#endif
 	}
     }
   else
-#endif
     {
       error (0, 0, _("%s has unknown file type"), quote (src_name));
       goto un_backup;
@@ -1627,6 +1718,9 @@ copy_internal (char const *src_name, char const *dst_name,
   if ( ! preserve_metadata)
     return true;
 
+  if (copied_as_regular)
+    return delayed_ok;
+
   /* POSIX says that `cp -p' must restore the following:
      - permission bits
      - setuid, setgid bits
@@ -1639,7 +1733,7 @@ copy_internal (char const *src_name, char const *dst_name,
      chown turns off set[ug]id bits for non-root,
      so do the chmod last.  */
 
-  if (!copied_as_regular && x->preserve_timestamps)
+  if (x->preserve_timestamps)
     {
       struct timespec timespec[2];
       timespec[0] = get_stat_atime (&src_sb);
@@ -1654,52 +1748,33 @@ copy_internal (char const *src_name, char const *dst_name,
     }
 
   /* Avoid calling chown if we know it's not necessary.  */
-  if (!(copied_as_regular && HAVE_FCHOWN) && x->preserve_ownership
+  if (x->preserve_ownership
       && (new_dst || !SAME_OWNER_AND_GROUP (src_sb, dst_sb)))
     {
-      if (chown (dst_name, src_sb.st_uid, src_sb.st_gid) == 0)
-	chown_succeeded = true;
-      else if (! chown_failure_ok (x))
-	{
-	  error (0, errno, _("failed to preserve ownership for %s"),
-		 quote (dst_name));
-	  if (x->require_preserve)
-	    return false;
-	}
+      if (! set_owner (x, dst_name, -1, src_sb.st_uid, src_sb.st_gid))
+	return false;
     }
 
-#if HAVE_STRUCT_STAT_ST_AUTHOR
-  /* Preserve the st_author field.  */
-  {
-    file_t file = file_name_lookup (dst_name, 0, 0);
-    if (file == MACH_PORT_NULL)
-      error (0, errno, _("failed to lookup file %s"), quote (dst_name));
-    else
-      {
-	error_t err = file_chauthor (file, src_sb.st_author);
-	if (err)
-	  error (0, err, _("failed to preserve authorship for %s"),
-		 quote (dst_name));
-	mach_port_deallocate (mach_task_self (), file);
-      }
-  }
-#endif
+  set_author (dst_name, -1, &src_sb);
 
-  /* Permissions of newly-created regular files are set by open and/or fchmod
-     in copy_reg.  But don't return early if there were any special bits and
-     chown succeeded, because the chown must have reset those bits.  */
-  if (copied_as_regular
-      && (HAVE_FCHMOD
-	  || (new_dst && !(chown_succeeded && (src_mode & ~S_IRWXUGO)))))
-    return delayed_ok;
-
-  if ((x->preserve_mode || new_dst)
-      && (x->copy_as_regular || S_ISREG (src_type) || S_ISDIR (src_type)))
+  if (x->preserve_mode || x->move_mode)
     {
-      if (chmod (dst_name, get_dest_mode (x, src_mode)) != 0)
+      if (copy_acl (src_name, -1, dst_name, -1, src_mode) != 0
+	  && x->require_preserve)
+	return false;
+    }
+  else if (x->set_mode)
+    {
+      if (set_acl (dst_name, -1, x->mode) != 0)
+	return false;
+    }
+  else if (restore_dst_mode)
+    {
+      if (lchmod (dst_name, dst_mode) != 0)
 	{
-	  error (0, errno, _("setting permissions for %s"), quote (dst_name));
-	  if (x->set_mode || x->require_preserve)
+	  error (0, errno, _("preserving permissions for %s"),
+		 quote (dst_name));
+	  if (x->require_preserve)
 	    return false;
 	}
     }
