@@ -1,5 +1,5 @@
 /* od -- dump files in octal and other formats
-   Copyright (C) 92, 1995-2005 Free Software Foundation, Inc.
+   Copyright (C) 92, 1995-2006 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -59,7 +59,7 @@ typedef double LONG_DOUBLE;
 # define LDBL_DIG DBL_DIG
 #endif
 
-#if HAVE_UNSIGNED_LONG_LONG
+#if HAVE_UNSIGNED_LONG_LONG_INT
 typedef unsigned long long int unsigned_long_long_int;
 #else
 /* This is just a place-holder to avoid a few `#if' directives.
@@ -337,7 +337,7 @@ All arguments to long options are mandatory for short options.\n\
       fputs (_("\
 \n\
 Traditional format specifications may be intermixed; they accumulate:\n\
-  -a   same as -t a,  select named characters\n\
+  -a   same as -t a,  select named characters, ignoring high-order bit\n\
   -b   same as -t o1, select octal bytes\n\
   -c   same as -t c,  select ASCII characters or backslash escapes\n\
   -d   same as -t u2, select unsigned decimal 2-byte units\n\
@@ -361,7 +361,7 @@ suffixes may be . for octal and b for multiply by 512.\n\
 \n\
 TYPE is made up of one or more of these specifications:\n\
 \n\
-  a          named character\n\
+  a          named character, ignoring high-order bit\n\
   c          ASCII character or backslash escape\n\
 "), stdout);
       fputs (_("\
@@ -497,7 +497,7 @@ dump_hexl_mode_trailer (size_t n_bytes, const char *block)
   for (i = n_bytes; i > 0; i--)
     {
       unsigned char c = *block++;
-      unsigned char c2 = (ISPRINT(c) ? c : '.');
+      unsigned char c2 = (isprint (c) ? c : '.');
       putchar (c2);
     }
   putchar ('<');
@@ -576,7 +576,7 @@ print_ascii (size_t n_bytes, void const *block,
 	  break;
 
 	default:
-	  sprintf (buf, (ISPRINT (c) ? "  %c" : "%03o"), c);
+	  sprintf (buf, (isprint (c) ? "  %c" : "%03o"), c);
 	  s = buf;
 	}
 
@@ -1469,7 +1469,7 @@ dump_strings (void)
 	      free (buf);
 	      return ok;
 	    }
-	  if (!ISPRINT (c))
+	  if (! isprint (c))
 	    /* Found a non-printing.  Try again starting with next char.  */
 	    goto tryline;
 	  buf[i] = c;
@@ -1492,7 +1492,7 @@ dump_strings (void)
 	    }
 	  if (c == '\0')
 	    break;		/* It is; print this string.  */
-	  if (!ISPRINT (c))
+	  if (! isprint (c))
 	    goto tryline;	/* It isn't; give up on this string.  */
 	  buf[i++] = c;		/* String continues; store it all.  */
 	}
@@ -1581,7 +1581,7 @@ main (int argc, char **argv)
   integral_type_size[sizeof (short int)] = SHORT;
   integral_type_size[sizeof (int)] = INT;
   integral_type_size[sizeof (long int)] = LONG;
-#if HAVE_UNSIGNED_LONG_LONG
+#if HAVE_UNSIGNED_LONG_LONG_INT
   /* If `long int' and `long long int' have the same size, it's fine
      to overwrite the entry for `long' with this one.  */
   integral_type_size[sizeof (unsigned_long_long_int)] = LONG_LONG;
